#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO
from dataclasses import dataclass
from datetime import datetime
from uuid import uuid4

import pytz
from pyecore.ecore import EDate

import esdl
from esdl.profiles.profilemanager import (
    ProfileManager,
    UnknownProfileNameException,
    ProfileType,
)
from influxdb import InfluxDBClient

from esdl.utils.datetime_utils import parse_date


@dataclass
class ConnectionSettings:
    """
    The ConnectionSettings class represents all required settings that are required to connect to an
    InfluxDB server
    """

    host: str
    """hostname of the InfluxDB server to connect to"""
    port: int
    """port to connect to"""
    username: str
    """username of the user that has permission to access the database"""
    password: str
    """password of the user"""
    database: str
    """database to connect to"""
    ssl: bool
    """indicates whether or not to use SSL when connecting to the InfluxDB server"""
    verify_ssl: bool
    """indicates wether or not to verify SSL certificates"""


class InfluxDBProfileManager(ProfileManager):
    """
    InfluxDBProfileManager: manages profile data that can be loaded from and saved to InfluxDB (v1 only).
    InfluxDBProfileManager is a subclass of ProfileManager, so it also provides functionality to convert from/to
    different ESDL profiles and to load/save to CSV
    """

    def __init__(self, settings: ConnectionSettings, source_profile=None):
        """
        Constructor of the InfluxDBProfileManager

        :param settings: connection settings to connect to the InfluxDB server
        :param source_profile: the source profile of which the data is copied into this profiles manager instance
        """
        super().__init__()

        # Temporary fix, for ESDL profiles that have HTTP(S) in the hostname
        settings.host = settings.host.replace("http://", "").replace("https://", "")

        self.database_settings = settings
        self.influxdb_client = InfluxDBClient(
            host=settings.host,
            port=settings.port,
            username=settings.username,
            password=settings.password,
            database=settings.database,
            ssl=settings.ssl,
            verify_ssl=settings.verify_ssl,
        )

        if settings.database not in [
            db["name"] for db in self.influxdb_client.get_list_database()
        ]:
            self.influxdb_client.create_database(settings.database)

        if source_profile:
            self.convert(source_profile)
            if isinstance(source_profile, InfluxDBProfileManager):
                self.database_settings = source_profile.database_settings

    def load_influxdb(
        self,
        measurement: str,
        fields: list,
        from_datetime: datetime = None,
        to_datetime: datetime = None,
        filters: list = None,
    ):
        """
        Loads profile information from InfluxDB

        :param measurement: the name of the measurement to use in InfluxDB
        :param fields: a list of field names that need to be loaded from InfluxDB
        :param from_datetime: the start datetime (included in the data)
        :param to_datetime: the end datetime (included in the data)
        :param filters: a list of dictionaries with 'tag' and 'value' keys, that can be used to filter on the data
        :return: None
        """
        self.clear_profile()

        where_clause_list = list()
        if from_datetime and to_datetime:
            influxdb_startdate = from_datetime.strftime("%Y-%m-%dT%H:%M:%SZ")
            influxdb_enddate = to_datetime.strftime("%Y-%m-%dT%H:%M:%SZ")

            where_clause_list.append("time >= '" + influxdb_startdate + "'")
            where_clause_list.append("time <= '" + influxdb_enddate + "'")

        if filters:
            for filter in filters:
                where_clause_list.append(f"{filter['tag']}='{filter['value']}'")

        fields_string = ",".join(['"' + field + '"' for field in fields])
        query_string = f'SELECT {fields_string} FROM "{measurement}"'
        if where_clause_list:
            query_string += " WHERE (" + " AND ".join(where_clause_list) + ")"

        res = self.influxdb_client.query(query_string)

        if res:
            res_list = res.get_points()
            for elem in res_list:
                dt = parse_date(elem["time"])

                try:
                    aware_dt = pytz.utc.localize(
                        dt
                    )  # Assume timezone is UTC if no TZ was given
                except (
                    ValueError
                ):  # ValueError: No naive datetime (tzinfo is already set)
                    aware_dt = dt

                row = [aware_dt]
                if self.start_datetime is None:
                    self.start_datetime = aware_dt

                for f in fields:
                    row.append(elem[f])

                self.profile_data_list.append(row)
                self.num_profile_items += 1

            self.end_datetime = self.profile_data_list[-1][0]
            header = ["datetime"]
            header.extend(fields)
            self.profile_header = header
            self.profile_type = ProfileType.DATETIME_LIST

            tags = None
            if filters:
                tags = dict()
                for filter in filters:
                    tags[filter["tag"]] = filter["value"]
            return self.get_esdl_influxdb_profile(measurement, fields, tags)
        else:
            return None

    @staticmethod
    def _parse_esdl_profile_filters(esdl_filters):
        if esdl_filters is None or esdl_filters == "":
            return []

        filter_str_list = esdl_filters.split(" AND ")
        filter_list = list()
        for f in filter_str_list:
            try:
                [key, value] = f.split("=")
                if key is not None and key != "" and value is not None and value != "":
                    filter_list.append({"key": key, "value": value})
                else:
                    raise WrongFilterFormatException(
                        "Filter specification in ESDL profile cannot be parsed"
                    )
            except:
                raise WrongFilterFormatException(
                    "Filter specification in ESDL profile cannot be parsed"
                )

        return filter_list

    @staticmethod
    def create_esdl_influxdb_profile_manager(
        esdl_profile: esdl.InfluxDBProfile,
        username: str = None,
        password: str = None,
        use_ssl: bool = False,
        verify_ssl: bool = False,
    ):
        """
        method to create an instance of InfluxDBProfileManager using an esdl.InfluxDBProfile. The data referenced to in
        the profile is loaded from InfluxDB

        :param esdl_profile: esdl.InfluxDBProfile that is used as an input profile
        :param username: username to be used to connect to the InfluxDB server mentioned in the esdl.InfluxDBProfile,
                         defaults to None
        :param password: password to be used to connect to the InfluxDB server mentioned in the esdl.InfluxDBProfile,
                         defaults to None
        :param use_ssl: indicates if HTTPS should be used instead of HTTP to connect to the InfluxDB server, defaults
                        to False
        :param verify_ssl: verify SSL certificates for HTTPS requests, defaults to False
        """
        conn_settings = ConnectionSettings(
            host=esdl_profile.host,
            port=esdl_profile.port,
            database=esdl_profile.database,
            username=username,
            password=password,
            ssl=use_ssl,
            verify_ssl=verify_ssl,
        )

        prof = InfluxDBProfileManager(conn_settings)
        prof.load_influxdb(
            measurement=esdl_profile.measurement,
            fields=[esdl_profile.field],
            from_datetime=esdl_profile.startDate,
            to_datetime=esdl_profile.endDate,
            filters=InfluxDBProfileManager._parse_esdl_profile_filters(
                esdl_profile.filters
            ),
        )

        return prof

    @staticmethod
    def _create_esdl_filters_from_tags_dict(tags: dict = None):
        if not tags:
            return ""
        else:
            filter_list = list()

            for k, v in tags.items():
                filter_list.append(f"{k}='{v}'")

            return " AND ".join(filter_list)

    def get_esdl_influxdb_profile(
        self, measurement: str, field_names: list, tags: dict = None
    ):
        """
        Creates an esdl.InfluxDBProfile instance (or a list of instances) that refers to the data in the database.
        Is called by load_influxdb and save_influxdb

        :param measurement: name of the InfluxDB measurement where the data must be written to
        :param field_names: list of the fields that should be written to InfluxDB
        :param tags: dictionary with tags and tag values, that should be used when writing this data to InfluxDB
        :return: an esdl.InfluxDBProfile instance or a list of esdl.InfluxDBProfile instances in case multiple fields
                 were specified, with proper references to the data in the database
        """
        profile_list = list()
        for field in field_names:
            esdl_profile = esdl.InfluxDBProfile(
                id=str(uuid4()),
                name=field,
                host=self.database_settings.host,
                port=self.database_settings.port,
                database=self.database_settings.database,
                measurement=measurement,
                field=field,
                filters=InfluxDBProfileManager._create_esdl_filters_from_tags_dict(
                    tags
                ),
                startDate=EDate.from_string(
                    self.start_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
                ),
                endDate=EDate.from_string(
                    self.end_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
                ),
            )
            profile_list.append(esdl_profile)

        if len(profile_list) == 1:
            return profile_list[0]
        else:
            return profile_list

    def save_influxdb(self, measurement: str, field_names: list, tags: dict = None):
        """
        Saves profile information to InfluxDB

        :param measurement: name of the InfluxDB measurement where the data must be written to
        :param field_names: list of the fields that should be written to InfluxDB
        :param tags: dictionary with tags and tag values, that should be used when writing this data to InfluxDB
        :return: an esdl.InfluxDBProfile instance or a list of esdl.InfluxDBProfile instances in case multiple fields
                 were specified, with proper references to the data in the database
        """
        json_body = []
        if not measurement or measurement == "":
            raise NoDataException(
                "Measurement name is not specified, so no data can be written to InfluxDB"
            )
        if not field_names:
            raise NoDataException(
                "No field names are specified, so no data can be written to InfluxDB"
            )
        if tags and not isinstance(tags, dict):
            raise WrongTagsFormatException(
                "The tags parameter should be a dictionary with tag names as keys"
            )
        for profile_row in self.profile_data_list:
            dt_string = profile_row[0].strftime("%Y-%m-%dT%H:%M:%S%z")
            fields = {}
            for fld_name in field_names:
                if fld_name in self.profile_header:
                    idx = self.profile_header.index(fld_name)
                    fields[self.profile_header[idx]] = profile_row[idx]
                else:
                    raise UnknownProfileNameException(
                        f"Given field name ({fld_name}) is not recognized as a profile name"
                    )

            if fields:
                data = {"measurement": measurement, "time": dt_string, "fields": fields}
                if tags:
                    data["tags"] = tags

                json_body.append(data)

        if json_body:
            self.influxdb_client.write_points(
                points=json_body,
                database=self.database_settings.database,
                batch_size=100,
            )
            return self.get_esdl_influxdb_profile(measurement, field_names, tags)

        else:
            return None


class NoDataException(Exception):
    """Thrown when no profile data can be written to InfluxDB"""

    pass


class WrongFilterFormatException(Exception):
    """Thrown when filter specification in ESDL profile cannot be parsed"""

    pass


class WrongTagsFormatException(Exception):
    """Thrown when the tags parameter has the wrong structure"""

    pass



import logging
from dataclasses import dataclass
from datetime import datetime
from logging import warning, error
from typing import Any, Optional, List, Tuple, Dict

import psycopg
from psycopg import Connection, sql

from esdl import esdl, QuantityAndUnitType
from esdl.profiles.data_configurations.credentials import Credentials
from esdl.units.parser import unit_to_string, build_qau_from_unit_string

# table name used for storing metadata that is not in the table
META_DATA_TABLE_NAME = "datatable_metadata"
META_DATA_TABLE_COLUMNS = [
    "table_name",
    "column_name",
    "physical_quantity",
    "unit",
    "profile_name",
]
META_DATA_TABLE_COLUMN_LENGTH = [64, 64, 32, 32, 255]  # VARCHAR lengths
META_DATA_TABLE_NOT_NULL = [True, True, True, False, False]

# set to True to enable debugging of SQL statements
DEBUG_SQL = False


@dataclass
class DataTableMetaData:
    qau: Optional[QuantityAndUnitType]
    name: Optional[str]


class PostgresqlConfiguration:
    """
    Implementation of PostgreSQL configuration for DataTableProfile
    DatabaseConfiguration(type=esdl.DatabaseTypeEnum.POSTGRESQL)
    """

    connection: Optional[Connection]
    datatable_profile: esdl.DataTableProfile

    def __init__(
        self,
        datatable_profile: esdl.DataTableProfile,
        credentials_dict: dict[str, Credentials],
    ):
        self.datatable_profile = datatable_profile
        self.connection: Optional[Connection] = None
        self.credentials_dict = credentials_dict
        self._connect_postgres(datatable_profile, credentials_dict)

    def _connect_postgres(
        self,
        datatable_profile: esdl.DataTableProfile,
        credentials_dict: dict[str, Credentials],
    ):
        try:
            configuration = datatable_profile.configuration
            # try on id of configuration
            credential = credentials_dict.get(configuration.id, None)
            if credential is None:
                # try on host name
                credential = credentials_dict.get(configuration.host, None)
                if credential is None:
                    raise InvalidCredentials(
                        f"DataConfiguration '{configuration.id}' is has no associated credentials"
                        f" to use for connecting to the Postgres database"
                    )

            if DEBUG_SQL:
                print(
                    f"Connecting to Postgres database at {credential.username}@{configuration.host}:"
                    f"{configuration.port or 5432}/{configuration.database}"
                )
            self.connection = psycopg.connect(
                user=credential.username,
                password=credential.password,
                host=configuration.host,
                port=configuration.port or 5432,
                dbname=configuration.database,
            )
        except Exception as e:
            print("Error while connecting to PostgreSQL: ", e)
            if self.connection and not self.connection.closed:
                self.connection.close()
                raise e
        return self.connection

    def get_connection(self):
        return self.connection

    def disconnect(self):
        if self.connection:
            self.connection.close()

    def load_data(
        self, additional_filters: Dict[str, Any] = None, column_based: bool = False
    ) -> Tuple[List[List[any]], List[str], List[DataTableMetaData | None]]:
        """
        Load data from the configured table using the DataTableProfile settings.
        Supports querying a single column or all columns. The datetime column is
        always returned as the first column.

        :param additional_filters: (key,value) set of additional filters for the query in the WHERE clause
        :param column_based: if True, data is returned by column instead of row
        :return: a tuple of profile_values[[]], header[], List[DataTableMetaData] with data loaded from Postgres.
        the datetime column is always returned as first column
        """
        if not self.connection:
            raise Exception("PostgreSQL connection not established")

        with self.connection.cursor() as cursor:
            if self.datatable_profile.schema:
                table_ident = sql.Identifier(
                    self.datatable_profile.schema, self.datatable_profile.tableName
                )
            else:
                table_ident = sql.Identifier(self.datatable_profile.tableName)

            dt_ident = sql.Identifier(self.datatable_profile.datetimeColumnName)
            # column is None if we want to query all columns
            col_ident = (
                sql.Identifier(self.datatable_profile.columnName)
                if self.datatable_profile.columnName
                else None
            )

            where_sql, params = PostgresqlConfiguration._build_where_clause(
                dt_ident,
                self.datatable_profile.startDate,
                self.datatable_profile.endDate,
                additional_filters,
            )

            if self.datatable_profile.columnName:
                query = sql.SQL("SELECT {dt}, {col} FROM {tbl} {where}").format(
                    dt=dt_ident, col=col_ident, tbl=table_ident, where=where_sql
                )
            else:
                # Fallback to select all columns
                query = sql.SQL("SELECT * FROM {tbl} {where}").format(
                    tbl=table_ident, where=where_sql
                )

            if DEBUG_SQL:
                print(query.as_string(cursor))

            cursor.execute(query, params)
            result = cursor.fetchall()

            # process results
            profile_values, header = PostgresqlConfiguration._process_query_results(
                cursor, result, self.datatable_profile.datetimeColumnName, column_based
            )
            metadata = [self.load_meta_data(c) for c in header]

            return profile_values, header, metadata

    def load_data_custom(
        self,
        schema: Optional[str],
        table_name: str,
        datetime_column_name: str,
        column_name: str,
        start_date: Optional[datetime],
        end_date: Optional[datetime],
        additional_filters: Optional[Dict[str, Any]],
        multiplier: Optional[float],
        downsample_bucket_sec: Optional[int],
        column_based: bool,
    ) -> Tuple[List[List[any]], List[str], List[DataTableMetaData | None]]:
        """
        Execute a flexible SQL query against a PostgreSQL table, supporting a SINGLE
        value column, datetime filtering, optional downsampling, and additional
        equality-based filters.

        :param schema: Optional schema name.
        :param table_name: Name of the table to query.
        :param datetime_column_name: Name of the datetime column.
        :param column_name: Name of the data column to query.
        :param start_date: Optional start datetime for filtering.
        :param end_date: Optional end datetime for filtering.
        :param additional_filters: Optional dict of column=value filters for the WHERE clause.
        :param multiplier: Optional multiplier for scaling profile value.
        :param downsample_bucket_sec: If provided, results are grouped into fixed-size
                                    time buckets and averaged.
        :param column_based: If True, return data grouped by column instead of row.

        :return: A tuple of (profile_values, header, metadata).
        """
        if not self.connection:
            raise Exception("PostgreSQL connection not established")

        with self.connection.cursor() as cursor:
            if schema:
                table_ident = sql.Identifier(schema, table_name)
            else:
                table_ident = sql.Identifier(table_name)

            dt_ident = sql.Identifier(datetime_column_name)
            col_ident = sql.Identifier(column_name)

            # Scale profile value if multiplier is given;
            # otherwise, DataTableProfile assigned multiplier or default value (1.0) is used.
            multiplier = (
                multiplier
                if multiplier is not None
                else self.datatable_profile.multiplier
            )

            where_sql, params = PostgresqlConfiguration._build_where_clause(
                dt_ident, start_date, end_date, additional_filters
            )

            # Downsampled query by grouping them into fixed-size buckets.
            if downsample_bucket_sec:
                query = sql.SQL(
                    """
                    SELECT
                        to_timestamp(
                            floor(extract(epoch from {dt}) / %s) * %s
                        )::timestamp AS {dt},
                        AVG({col} * %s) AS {col}
                    FROM {tbl}
                    {where}
                    GROUP BY 1
                    ORDER BY 1
                """
                ).format(dt=dt_ident, col=col_ident, tbl=table_ident, where=where_sql)
                params = [
                    downsample_bucket_sec,
                    downsample_bucket_sec,
                    multiplier,
                    *params,
                ]

            # Fallback to normal query
            else:
                query = sql.SQL(
                    "SELECT {dt}, ({col} * %s) AS {col} FROM {tbl} {where}"
                ).format(dt=dt_ident, col=col_ident, tbl=table_ident, where=where_sql)
                params = [multiplier, *params]

            if DEBUG_SQL:
                print(query.as_string(cursor))

            cursor.execute(query, params)
            result = cursor.fetchall()

            profile_values, header = PostgresqlConfiguration._process_query_results(
                cursor, result, datetime_column_name, column_based
            )

            metadata = [self.load_meta_data(c) for c in header]

            return profile_values, header, metadata

    @staticmethod
    def _build_where_clause(
        dt_ident,
        start_date: Optional[datetime],
        end_date: Optional[datetime],
        additional_filters: Dict[str, Any] = None,
    ) -> Tuple[Any, List]:
        """
        A utility function to build a SQL WHERE clause for datetime range and simple equality filters.

        :param dt_ident: psycopg2 SQL identifier for the datetime column.
        :param start_date: Optional start datetime for filtering.
        :param end_date: Optional end datetime for filtering.
        :param additional_filters: Optional dict of column=value filters.
        :return: A tuple of (where_sql, params_list).
        """
        if additional_filters is not None and not isinstance(additional_filters, dict):
            raise TypeError("additional_filters must be a dict or None")

        where_clauses = []
        params = []

        dt_str_format = "%Y-%m-%dT%H:%M:%SZ"

        if start_date:
            where_clauses.append(sql.SQL("{dt} >= %s").format(dt=dt_ident))
            params.append(start_date.strftime(dt_str_format))

        if end_date:
            where_clauses.append(sql.SQL("{dt} <= %s").format(dt=dt_ident))
            params.append(end_date.strftime(dt_str_format))

        # NOTE: 'OR' or more complex filtering is not supported yet.
        if additional_filters:
            for col, val in additional_filters.items():
                where_clauses.append(
                    sql.SQL("{col} = %s").format(col=sql.Identifier(col))
                )
                params.append(val)

        where_sql = (
            sql.SQL(" WHERE ") + sql.SQL(" AND ").join(where_clauses)
            if where_clauses
            else sql.SQL("")
        )

        return where_sql, params

    @staticmethod
    def _process_query_results(
        cursor, result: List, datetime_column_name: str, column_based: bool
    ) -> Tuple[List[List], List[str]]:
        """
        Reorder results so the datetime column is always first.
        Supports row-based or column-based output.
        """
        columns = cursor.description
        column_names = [c.name for c in columns]

        try:
            dt_index = column_names.index(datetime_column_name)
        except ValueError:
            dt_index = 0

        dt_col = column_names.pop(dt_index)
        header = [dt_col, *column_names]

        profile_values = [[] for _ in range(len(header))] if column_based else []

        for row in result:
            row_list = list(row)
            dt_val = row_list.pop(dt_index)
            ordered = [dt_val, *row_list]

            if not column_based:
                profile_values.append(ordered)
            else:
                for i, val in enumerate(ordered):
                    profile_values[i].append(val)

        return profile_values, header

    def load_meta_data(self, column_name: str = None) -> None | DataTableMetaData:
        """
        Loads metadata from meta_data table

        :return: configures
        """
        if not self.connection:
            raise Exception("PostgreSQL connection not established")
        with self.connection.cursor() as cursor:
            if self.datatable_profile.schema:
                table = sql.Identifier(
                    self.datatable_profile.schema, META_DATA_TABLE_NAME
                )
            else:
                table = sql.Identifier(META_DATA_TABLE_NAME)

            columns = [sql.Identifier(c) for c in META_DATA_TABLE_COLUMNS]
            query = sql.SQL(
                "SELECT {fields} FROM {table} WHERE table_name=%s AND column_name=%s"
            ).format(fields=sql.SQL(", ").join(columns), table=table)
            if DEBUG_SQL:
                print(query.as_string(cursor))

            if self.datatable_profile.columnName:
                column_name = self.datatable_profile.columnName
            try:
                cursor.execute(query, (self.datatable_profile.tableName, column_name))
                result = cursor.fetchone()
                if result:
                    result_dict = dict(zip(META_DATA_TABLE_COLUMNS, result))
                    if result_dict["unit"] or result_dict["physical_quantity"]:
                        qau = build_qau_from_unit_string(
                            result_dict["unit"], result_dict["physical_quantity"]
                        )
                        if DEBUG_SQL:
                            print(
                                f'QaU metadata found: {result_dict["physical_quantity"]} in {result_dict["unit"]}'
                            )
                        self.datatable_profile.profileQuantityAndUnit = qau
                    if result_dict["profile_name"]:
                        self.datatable_profile.name = result_dict["profile_name"]
                    return DataTableMetaData(qau, result_dict["profile_name"])
                else:
                    warning(
                        f"No metadata found for table {table.as_string()} and column '{column_name or self.datatable_profile.columnName}'"
                        f" in database '{self.datatable_profile.configuration.database}', cannot derive unit"
                    )
                    return None
            except Exception as e:
                warning(
                    f"No metadata found for table {table.as_string()} and column '{column_name or self.datatable_profile.columnName}'"
                    f" in database '{self.datatable_profile.configuration.database}', cannot derive units"
                )

    def save_data(
        self, profile_values: List[List[Any]], header: List[str], overwrite: bool = True
    ):
        """
        Saves data to Postgres. It will insert all the columns in the profile_values by default.

        :param profile_values:
        :param header:
        :param overwrite: If True, truncate the existing table rows and insert new ones.
        :return:
        """
        if not self.connection:
            raise Exception("PostgreSQL connection not established")
        if not self.datatable_profile.tableName:
            raise InvalidDataTableProfile(
                f"Missing tableName in DataTableProfile with id={self.datatable_profile.id}"
            )

        if self.datatable_profile.schema:
            table = sql.Identifier(
                self.datatable_profile.schema, self.datatable_profile.tableName
            )
        else:
            table = sql.Identifier(self.datatable_profile.tableName)

        # COMMENT ON COLUMN essim_run_20250804.csv_data_table.column2 IS 'unit=PJ';

        with self.connection.cursor() as cursor:
            non_datetime_columns = list(header)
            try:
                non_datetime_columns.remove(self.datatable_profile.datetimeColumnName)
            except ValueError as e:
                logging.error(
                    f"Cannot find datetime column '{self.datatable_profile.datetimeColumnName}' in dataset, "
                )
                logging.error(
                    f"configure using esdl.DataTableProfile.datetimeColumnName"
                )
                raise e
            insert_columns = [f'"{c}" DOUBLE PRECISION' for c in non_datetime_columns]
            insert_columns.insert(
                0, f'"{self.datatable_profile.datetimeColumnName}" TIMESTAMP'
            )
            insert_column_statement = ",\n".join(insert_columns)

            # create schema and table if not exists
            if self.datatable_profile.schema:
                query = sql.SQL("CREATE SCHEMA IF NOT EXISTS {schema}").format(
                    schema=sql.Identifier(self.datatable_profile.schema)
                )
                if DEBUG_SQL:
                    print(query.as_string(cursor))
                cursor.execute(query)

            sql_string = (
                "CREATE TABLE IF NOT EXISTS {table}" + f" ({insert_column_statement})"
            )
            query = sql.SQL(sql_string).format(table=table)
            if DEBUG_SQL:
                print(query.as_string(cursor))
            cursor.execute(query)
            # create index on table datetime column
            sql_string = r"CREATE INDEX IF NOT EXISTS {table}_metadata_datetime_idx ON {table} USING btree ({column_name})"
            query = sql.SQL(sql_string).format(
                table=table,
                column_name=sql.Identifier(self.datatable_profile.datetimeColumnName),
            )
            cursor.execute(query)
            self.connection.commit()

            if overwrite:
                query = sql.SQL("TRUNCATE TABLE {table}").format(table=table)
                if DEBUG_SQL:
                    print(query.as_string(cursor))
                cursor.execute(query)

            columns = sql.SQL(", ").join(sql.Identifier(h) for h in header)
            query = sql.SQL("COPY {table} ({columns}) FROM STDIN").format(
                table=table, columns=columns
            )
            if DEBUG_SQL:
                print(query.as_string(cursor))
            with cursor.copy(query) as copy:
                for row in profile_values:
                    copy.write_row(row)

            # Set extra information in tableName_metadata table (in same schema)
            # such as Unit, Quantity, multiplier, and name of profile (and id?)

            for column_name in header:
                self.save_meta_data(self.datatable_profile, column_name)
            self.connection.commit()

    def save_meta_data(
        self, datatable_profile: esdl.DataTableProfile, column_name: str = None
    ):
        """
        Meta data format in schema
        | table name | column_name | Unit | Quantity | description |

        :param datatable_profile: the data table profile including unit information and description
        :param column_name: [optional] the column name if not set in the datatable profile
                (when inserting multiple columns)
        :return: None
        """
        qau = datatable_profile.profileQuantityAndUnit
        if qau is None or qau.unit is None or qau.physicalQuantity is None:
            raise InvalidDataTableProfile(
                f"Missing Quantity and unit information for DataTableProfile with id={datatable_profile.id}"
            )
        if column_name == datatable_profile.datetimeColumnName:
            qau = QuantityAndUnitType(
                id="time",
                unit=esdl.UnitEnum.NONE,
                physicalQuantity=esdl.PhysicalQuantityEnum.TIME,
            )

        if self.datatable_profile.schema:
            table = sql.Identifier(self.datatable_profile.schema, META_DATA_TABLE_NAME)
        else:
            table = sql.Identifier(META_DATA_TABLE_NAME)

        # META_DATA_TABLE_COLUMNS = ["table_name", "column_name",  "quantity", "unit", "description"]
        insert_columns = [
            f"{c} VARCHAR({n}){' NOT NULL' if nn else ''}"
            for c, n, nn in zip(
                META_DATA_TABLE_COLUMNS,
                META_DATA_TABLE_COLUMN_LENGTH,
                META_DATA_TABLE_NOT_NULL,
            )
        ]
        insert_column_statement = ",\n".join(insert_columns)
        sql_string = (
            "CREATE TABLE IF NOT EXISTS {table}"
            + f" ({insert_column_statement}"
            + f", PRIMARY KEY ({META_DATA_TABLE_COLUMNS[0]},{META_DATA_TABLE_COLUMNS[1]}))"
        )
        with self.connection.cursor() as cursor:
            # create metadata table if not exists
            query = sql.SQL(sql_string).format(table=table)
            if DEBUG_SQL:
                print(query.as_string(cursor))
            cursor.execute(query)

            # insert data if not exists
            upsert_sql = (
                "INSERT INTO {table} "
                + f"({','.join(META_DATA_TABLE_COLUMNS)}) VALUES "
                + f"({','.join( ['%s'] * len(META_DATA_TABLE_COLUMNS) )})"
                + f" ON CONFLICT ({','.join(META_DATA_TABLE_COLUMNS[:2])} ) DO UPDATE SET ({','.join(META_DATA_TABLE_COLUMNS[2:])}) = "
                + "("
                + ",".join(
                    [f"EXCLUDED.{column}" for column in META_DATA_TABLE_COLUMNS[2:]]
                )
                + ")"
            )

            query = sql.SQL(upsert_sql).format(table=table)
            if DEBUG_SQL:
                print(query.as_string(cursor))
            unit_string = unit_to_string(qau)
            if datatable_profile.columnName:
                insert_column_name = datatable_profile.columnName
            else:
                insert_column_name = column_name
            try:
                cursor.execute(
                    query,
                    (
                        datatable_profile.tableName,
                        insert_column_name,
                        qau.physicalQuantity.name,
                        unit_string,
                        datatable_profile.name,
                    ),
                )
            except Exception as e:
                error("Error", e)


class InvalidCredentials(Exception):
    """Thrown when no credentials are provided for connecting to PostgreSQL."""

    pass


class InvalidDataTableProfile(Exception):
    """Thrown when the DataTableProfile is not complete"""

    pass



from dataclasses import dataclass


@dataclass
class Credentials:
    """
    Simple dataclass that contains username and password used for logging into a database

    Use the static function create_dict() to create a compatible dict to be used in the DataTableProfileManager
    load() and save() methods for authentication
    """

    username: str
    password: str

    @staticmethod
    def create_dict(id: str, username: str, password: str) -> dict[str, "Credentials"]:
        """
        Creates a new credential dict with a mapping from id -> Credential[Username, Password] which can be used directly
        in the load() and save() methods of the DataTableProfileManager class
        :param id: refers to the id of the DatabaseConfiguration for which these credentials should be used. Alternatively
                    you can use the hostname as id, for more generic matching.
        :param username: Username
        :param password: Password
        :return: dict[str, 'Credentials']
        """
        d = dict()
        d[id] = Credentials(username, password)
        return d



#  This work is based on original code developed and copyrighted by TNO 2023.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO
from os import environ
from typing import Dict, List, Any, Tuple, Optional
from datetime import datetime
from uuid import uuid4

from pyecore.ecore import EDate

import esdl
from esdl import DataTableProfile
from esdl.profiles.data_configurations.credentials import Credentials
from esdl.profiles.data_configurations.postgresql import (
    PostgresqlConfiguration,
    DataTableMetaData,
)
from esdl.profiles.excelprofilemanager import ExcelProfileManager
from esdl.profiles.profilemanager import (
    ProfileManager,
    ProfileType,
    NoProfileLoadedExecption,
)

global_credentials: Dict[str, Credentials] = {}

# add a default credential based on hostname based on environmental variables
if environ.get("DB_HOST", None):
    host = environ.get("DB_HOST", "localhost")
    global_credentials[host] = Credentials(
        username=environ.get("DB_USER", None), password=environ.get("DB_PASSWORD", None)
    )
    print(f"Detected DB credentials for {host}")


class DataTableProfileManager(ProfileManager):
    """
    DataTableProfileManager: manages profile data that can be loaded from and saved to Postgres or InfluxDB.
    DataTableProfileManager is a subclass of ProfileManager, so it also provides functionality to convert from/to
    different ESDL profiles and to load/save to CSV
    """

    data_table_profile: esdl.DataTableProfile
    metadata: List[DataTableMetaData]

    def __init__(
        self, data_table_profile: esdl.DataTableProfile = None, source_profile=None
    ):
        """
        Constructor of the DataTableProfileManager

        :param data_table_profile: the DataTableProfile to load
        :param source_profile: the source profile of which the data is copied into this profiles manager instance
        """
        super().__init__()
        self.metadata = []
        self.profile_type = ProfileType.DATETIME_LIST

        if data_table_profile:
            self.data_table_profile = data_table_profile
        else:
            self.data_table_profile = esdl.DataTableProfile()

        if source_profile:
            self.convert(source_profile)

    def add_credential(self, credentials_dict: Dict[str, Credentials]):
        global_credentials.update(credentials_dict)

    def set_data_table_profile(self, data_table_profile: esdl.DataTableProfile):
        self.data_table_profile = data_table_profile

    def load_database_configuration(
        self,
        configuration: esdl.DatabaseConfiguration = None,
        credentials_dict: dict[str, Credentials] = None,
    ):
        """
        Loads profile data from a database configuration into
        profile_info and profile_header

        :param configuration: the database configuration, optional, otherwise
                            the one provided in the constructor is used
        :param credentials_dict: optional credentials dict to use to connect to the database
        :return: None
        """
        if not configuration:
            configuration = self.data_table_profile.configuration
        if not configuration:
            raise InvalidDataConfiguration("Missing configuration of DataTableProfile")

        if not credentials_dict:
            credentials_dict = global_credentials

        self.clear_profile()
        self.profile_type = ProfileType.DATETIME_LIST

        if configuration.type == esdl.DatabaseTypeEnum.POSTGRESQL:
            # handle postgres
            postgres = PostgresqlConfiguration(
                self.data_table_profile, credentials_dict
            )
            self.profile_data_list, self.profile_header, self.metadata = (
                postgres.load_data()
            )
            postgres.disconnect()
            if len(self.profile_data_list) > 0:
                self.num_profile_items = len(self.profile_data_list[0])
                if self.num_profile_items > 0:
                    dt_index = self.get_profile_name_index(
                        self.data_table_profile.datetimeColumnName
                    )
                    self.start_datetime = self.profile_data_list[0][dt_index]
                    self.end_datetime = self.profile_data_list[-1][dt_index]
        else:
            raise UnsupportedDataConfiguration(
                f"DataConfiguration of type {configuration.type.name} is not yet supported"
            )

    def get_profile_with_multiplier(
        self, column_based: bool = False
    ) -> List[List[float]]:
        """
        Return the loaded profile data with the configured multiplier applied to a column or all
        columns (if columnName is not set). The datetime column is returned unchanged as the first
        column.

        :param column_based: If True, return data grouped by column instead of row. If False, return row-based data.
        :return: A list of columns, where the first column is the datetime values and
                subsequent columns are scaled numeric data columns.
        :raises NoProfileLoadedExecption: If no profile data has been loaded.
        """

        if len(self.profile_header) == 0:
            raise NoProfileLoadedExecption("No profile loaded, header is empty.")

        # Default to 1.0 if not explicitly set.
        multiplier = self.data_table_profile.multiplier

        # Default to 'datetime' if not explicitly set.
        dt_col_name = self.data_table_profile.datetimeColumnName
        dt_index = self.profile_header.index(dt_col_name)

        if self.data_table_profile.columnName is not None:
            # A single column case
            col_names = [self.data_table_profile.columnName]
        else:
            # Multi-column mode: all columns except datetime
            col_names = [c for c in self.profile_header if c != dt_col_name]

        scaled_profile_data = []

        # Process row-based structure
        for row in self.profile_data_list:
            dt_value = row[dt_index]
            scaled_row = [dt_value]

            for col_name in col_names:
                col_index = self.profile_header.index(col_name)
                value = row[col_index]
                scaled_row.append(value * multiplier)

            scaled_profile_data.append(scaled_row)

        # Convert to column-based if requested
        if column_based:
            return [list(col) for col in zip(*scaled_profile_data)]

        return scaled_profile_data

    @staticmethod
    def load(
        data_table_profile: esdl.DataTableProfile,
        credentials_dict: dict[str, Credentials] = None,
    ) -> "DataTableProfileManager":
        """
        Static method to create an instance of DataTableProfileManager using an esdl.DataTableProfile. The data
        referenced to in the profile is loaded from the associated DataConfiguration
        (e.g. DatabaseConfiguration or FileConfiguration)

        :param data_table_profile: esdl.DataTableProfile that is used as an input profile
        :param credentials_dict: dictionary of credentials to use for authentication with ['id' -> Credentials] mapping,
         where 'id' refers to the id of the AbstractDataConfiguration that belongs to these credentials

        Note:
        - for FileConfigurations: uri without extension is used as tableName for CSV files if no tableName is given
        """

        if not credentials_dict:
            # use global credentials list
            credentials_dict = global_credentials

        configuration = data_table_profile.configuration

        if configuration is None:
            raise InvalidDataConfiguration("No DataConfiguration provided")
        if isinstance(configuration, esdl.DatabaseConfiguration):
            dtpman = DataTableProfileManager(data_table_profile)
            dtpman.load_database_configuration(configuration, credentials_dict)
            return dtpman
        elif isinstance(configuration, esdl.FileConfiguration):
            if configuration.type == esdl.FileTypeEnum.EXCEL:
                epm = ExcelProfileManager()
                # tableName is used as sheet name (see ecore model documentation)
                # TODO: load_excel ignores currently datetimeColumnName and columnName
                epm.load_excel(configuration.uri, data_table_profile.tableName)
                dtpman = DataTableProfileManager(data_table_profile, epm)
                return dtpman
            elif configuration.type == esdl.FileTypeEnum.CSV:
                dtpman = DataTableProfileManager(data_table_profile)
                dtpman.load_csv(configuration.uri)
                if not data_table_profile.tableName:
                    data_table_profile.tableName = (
                        configuration.uri.split(".")[0]
                        if configuration.uri
                        else "csv_file"
                    )
                    data_table_profile.tableName = data_table_profile.tableName.split(
                        "/"
                    )[-1]
                    data_table_profile.tableName = data_table_profile.tableName.split(
                        "\\"
                    )[-1]
                return dtpman
            else:
                raise UnsupportedDataConfiguration(
                    f"DataConfiguration of type {configuration.type.name} is not yet supported"
                )
        else:
            raise UnsupportedDataConfiguration(
                f"DataConfiguration {configuration} is not yet supported"
            )

    @staticmethod
    def query(
        data_table_profile: esdl.DataTableProfile,
        credentials_dict: Optional[Dict[str, Credentials]],
        table_name: str,
        column_name: str,
        schema: Optional[str] = None,
        datetime_column_name: Optional[str] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        additional_filters: Optional[Dict[str, Any]] = None,
        multiplier: Optional[float] = None,
        downsample_bucket_sec: Optional[int] = None,
        column_based: bool = False,
    ) -> Tuple[List[List[float]], List[str], List[esdl.QuantityAndUnitType]]:
        """
        Query data from a PostgreSQL-backed DataTableProfile using the associated
        DatabaseConfiguration. This method acts as the high-level API for retrieving
        profile data, applying optional filtering, scaling, and downsampling.

        Data is NOT stored inside this class, but passed directly
        to the caller.

        :param data_table_profile: The esdl.DataTableProfile describing the data source and default query parameters.
        :param credentials_dict: Optional mapping of configuration ID → Credentials. If None, global credentials are used.
        :param table_name: Name of the database table to query.
        :param column_name: Name of the single value column to retrieve.
        :param schema: Optional schema name. If None, the default schema is used.
        :param datetime_column_name: Optional override for the datetime column name. If None, the profile's datetime column is used.
        :param start_date: Optional start datetime for filtering.
        :param end_date: Optional end datetime for filtering.
        :param additional_filters: Optional (key, value) set of additional filters applied in the SQL WHERE clause.
        :param multiplier: Optional scaling factor applied to the value column. If None, the profile's default multiplier is used.
        :param downsample_bucket_sec: Optional bucket size (in seconds) for time-based downsampling using AVG().
        :param column_based: If True, results are returned column-based instead of row-based.

        :return: Tuple with: List of data (column-based if configured), header names,
                QuantityAndUnit if available for each header
        """
        if not credentials_dict:
            # use global credentials list
            credentials_dict = global_credentials

        configuration = data_table_profile.configuration
        if isinstance(configuration, esdl.DatabaseConfiguration):
            if configuration.type == esdl.DatabaseTypeEnum.POSTGRESQL:
                postgres = PostgresqlConfiguration(data_table_profile, credentials_dict)

                dt_column_name = (
                    datetime_column_name
                    if datetime_column_name is not None
                    else data_table_profile.datetimeColumnName
                )

                profile_data_list, profile_header, metadata = postgres.load_data_custom(
                    schema=schema,
                    table_name=table_name,
                    datetime_column_name=dt_column_name,
                    column_name=column_name,
                    start_date=start_date,
                    end_date=end_date,
                    additional_filters=additional_filters,
                    multiplier=multiplier,
                    downsample_bucket_sec=downsample_bucket_sec,
                    column_based=column_based,
                )
                qaus = [m.qau for m in metadata]
                postgres.disconnect()
                return profile_data_list, profile_header, qaus
            elif configuration.type == esdl.DatabaseTypeEnum.INFLUXDB:
                # TODO: support Influxdb
                pass
            else:
                raise UnsupportedDataConfiguration(
                    f"DataConfiguration of type {configuration.type.name} "
                    f"is not yet supported for direct querying"
                )
        else:
            raise UnsupportedDataConfiguration(
                f"DataConfiguration of type {configuration.type.name} "
                f"is not yet supported for direct querying"
            )

    def get_data_table_profile(
        self, profile_name: str = None, table_name: str = None, schema_name: str = None
    ) -> DataTableProfile | list[DataTableProfile]:
        """Creates an esdl.DataTableProfile instance that refers to the data loaded in this ProfileManager

        :param profile_name: name of the profile (one of the columns retrieved from the database), if None all will be returned
        :param table_name: name of the table, if None the value of self.data_table_profile.tableName will be used
        :param schema_name: name of the schema, if required. Defaults to None
        """

        if len(self.profile_header) == 0:
            raise NoProfileLoadedExecption("No profile loaded, header is empty.")

        datetime_columnname = self.profile_header[0]
        if table_name is None:
            table_name = self.data_table_profile.tableName

        if profile_name:  # return only one
            profile_name_index = self.get_profile_name_index(profile_name)
            profile_name = self.profile_header[profile_name_index]

            self.data_table_profile = esdl.DataTableProfile(
                id=str(uuid4()),
                name=profile_name,
                tableName=table_name,
                datetimeColumnName=datetime_columnname,
                columnName=profile_name,
                schema=schema_name,
                startDate=EDate.from_string(
                    self.start_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
                ),
                endDate=EDate.from_string(
                    self.end_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
                ),
            )
            if self.metadata and self.metadata[profile_name_index]:
                self.data_table_profile.profileQuantityAndUnit = self.metadata[
                    profile_name_index
                ].qau
                self.data_table_profile.name = self.metadata[profile_name_index].name
            return self.data_table_profile
        else:
            dtp_list = []
            for i, column in enumerate(self.profile_header[1:], start=1):
                dtp = esdl.DataTableProfile(
                    id=str(uuid4()),
                    name=profile_name,
                    tableName=table_name,
                    datetimeColumnName=datetime_columnname,
                    columnName=column,
                    schema=schema_name,
                    startDate=EDate.from_string(
                        self.start_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
                    ),
                    endDate=EDate.from_string(
                        self.end_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
                    ),
                )
                dtp_list.append(dtp)
                if self.metadata and self.metadata[i]:
                    dtp.profileQuantityAndUnit = self.metadata[i].qau
                    dtp.name = self.metadata[i].name
            return dtp_list

    # def save_tags(self, table: str, column_names: list, schema: str = None, tags: dict = None):
    #     """
    #     Saves profile information to InfluxDB
    #
    #     :param table: name of the table where the data must be written to. Table is created when not existing
    #     :param column_names: list of the columns that should be written
    #     :param tags: dictionary with tags and tag values, that should be used when writing this data in a JSON column
    #                 in the database
    #     :param schema: name of the schema, if required.
    #     :return: an esdl.DataTableProfile instance or a list of esdl.DataTableProfile instances in case multiple fields
    #              were specified, with proper references to the data in the database
    #     """
    #     pass

    def save(self, credentials_dict: dict[str, Credentials] = None, drop=True):
        """
        Saves the current datatable profile to the database

        :param credentials_dict: Dict with Credentials to connect to the database, or use global defined credentials
        :param drop: Drop data in the referenced table before writing data (default=True)
        """

        if not credentials_dict:
            credentials_dict = global_credentials

        if (
            self.data_table_profile.configuration.type
            == esdl.DatabaseTypeEnum.POSTGRESQL
        ):
            # handle postgres
            postgres = PostgresqlConfiguration(
                self.data_table_profile, credentials_dict
            )
            postgres.save_data(
                self.profile_data_list, self.profile_header, overwrite=drop
            )
            postgres.disconnect()
        else:
            raise UnsupportedDataConfiguration(
                f"DataConfiguration {self.data_table_profile.configuration.type.name} is not (yet) supported"
            )


class NoDataException(Exception):
    """Thrown when no profile data can be written to Postgres"""

    pass


class WrongFilterFormatException(Exception):
    """Thrown when filter specification in ESDL profile cannot be parsed"""

    pass


class InvalidDataConfiguration(Exception):
    """Thrown when there is no or invalid DataConfiguration"""

    pass


class UnsupportedDataConfiguration(Exception):
    """Thrown when this DataConfiguration is not supported"""

    pass



#  This work is based on original code developed and copyrighted by TNO 2020.
#  Subsequent contributions are licensed to you by the developers of such code and are
#  made available to the Project under one or several contributor license agreements.
#
#  This work is licensed to you under the Apache License, Version 2.0.
#  You may obtain a copy of the license at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#      TNO         - Initial implementation
#  Manager:
#      TNO

"""
Support functions for managing EObjects
"""
from typing import Dict

from pyecore.ecore import EAttribute, EObject, EClass, EReference, EStructuralFeature
from pyecore.innerutils import InternalSet
from pyecore.valuecontainer import ECollection
import logging

logger = logging.getLogger(__name__)



# add support for shallow copying or cloning an object
# it copies the object's attributes (e.g. clone an object), does only shallow copying
def clone(self):
    """
    Shallow copying or cloning an object
    It only copies the object's attributes (e.g. clone an object)
    Usage object.clone() or copy.copy(object) (as __copy__() is also implemented)
    :param self:
    :return: A clone of the object
    """
    newone = type(self)()
    eclass = self.eClass
    for x in eclass.eAllStructuralFeatures():
        if isinstance(x, EAttribute):
            #logger.trace("clone: processing attribute {}".format(x.name))
            if x.many:
                eOrderedSet = newone.eGet(x.name)
                for v in self.eGet(x.name):
                    eOrderedSet.append(v)
            else:
                newone.eSet(x.name, self.eGet(x.name))
    return newone


def deepcopy(self, memo=None, uuid_dict: Dict[str, EObject]=None, target_es: EObject=None, copy_xrefs=True) -> EObject:
    """
    Deep copying an EObject.

    :param memo: is used in recursive calls to find out which object have already been cloned
    :param uuid_dict: can be passed by the user to update the uuid_dict of the resource while deepcopying
    :param target_es search for cross-references in the target_es energy system or eObject if not found in copy.
           This can be used if the source ESDL has the same IDs as the target ESDL (e.g. quantity and units,
           or carriers) to make use the cross references are also connected in the deepcopy, before adding it to the
           `target_es`
    :param copy_xrefs copy cross-references (default). Set to False to not copy cross-references, e.g. in case
           you want to avoid copying references to other ESDLs that not have them. E.g. when cutting out an asset
           from an ESDL, without references to other parts of this ESDL.
    :return a deepcopy of the object.
    """
    first_call = False
    if memo is None:
        memo = dict()
        first_call = True
    if self in memo:
        return memo[self]

    copy: EObject = self.clone()
    if uuid_dict is not None and hasattr(copy, 'id'):
        uuid_dict[copy.id] = copy
    eclass: EClass = self.eClass
    for x in eclass.eAllStructuralFeatures():
        if isinstance(x, EReference):
            ref: EReference = x
            value: EStructuralFeature = self.eGet(ref)
            if value is None:
                continue
            if ref.containment:
                if ref.many and isinstance(value, ECollection):
                    #clone all containment elements
                    eAbstractSet: ECollection = copy.eGet(ref.name)
                    for ref_value in value:
                        duplicate = ref_value.__deepcopy__(memo, uuid_dict, target_es, copy_xrefs)
                        eAbstractSet.append(duplicate)
                else:
                    copy.eSet(ref.name, value.__deepcopy__(memo, uuid_dict, target_es, copy_xrefs))
            #else:
            #    # no containment relation, but a reference
            #    # this can only be done after a full copy
            #    pass
    # now copy should a full copy, but without cross-references

    memo[self] = copy

    if first_call and copy_xrefs:
        #logger.debug("copying references")
        for k, v in memo.items():
            eclass: EClass = k.eClass
            for x in eclass.eAllStructuralFeatures():
                if isinstance(x, EReference):
                    #logger.debug("deepcopy: processing x-reference {}".format(x.name))
                    ref: EReference = x
                    orig_value: EStructuralFeature = k.eGet(ref)
                    if orig_value is None:
                        continue
                    if not ref.containment:
                        opposite = ref.eOpposite
                        if opposite and opposite.containment:
                            # do not handle eOpposite relations, they are handled automatically in pyEcore
                            continue
                        if x.many:
                            eAbstractSet = v.eGet(ref.name)
                            for orig_ref_value in orig_value:
                                copy_ref_value = None
                                try:
                                    copy_ref_value = memo[orig_ref_value]
                                except KeyError:
                                    try:
                                        if target_es:
                                            if hasattr(orig_ref_value, 'id'): # use id to find object in target_es
                                                r = target_es.eResource
                                                ref_id = orig_ref_value.id
                                                copy_ref_value = r.uuid_dict[ref_id]
                                                #logger.warning(
                                                #    f'Using reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using {copy_ref_value}')
                                            else:
                                                logger.warning(
                                                    f'deepcopy(): No id found in target_es, cannot find reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                                copy_ref_value = orig_ref_value
                                        else:
                                            logger.warning(
                                                f'deepcopy(): Cannot find reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                            copy_ref_value = orig_ref_value
                                    except Exception:
                                        logger.warning(f'deepcopy(): Cannot find reference of type {orig_ref_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                        copy_ref_value = orig_ref_value
                                eAbstractSet.append(copy_ref_value)
                        else:
                            try:
                                copy_value = memo[orig_value]
                            except KeyError:
                                try:
                                    if target_es:
                                        if hasattr(orig_value, 'id'):  # use id to find object in target_es
                                            r = target_es.eResource
                                            ref_id = orig_value.id
                                            copy_value = r.uuid_dict[ref_id]
                                            # logger.debug(
                                            #     f'Using reference of type {orig_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using {copy_value}')
                                        else:
                                            logger.warning(
                                                f'deepcopy(): Referenced object has no ID, Can\'t find reference for {orig_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo or target_es, using original {orig_value}')
                                            copy_value = orig_value
                                    else:
                                        logger.warning(
                                            f'deepcopy(): Can\'t find reference for {orig_value.eClass.name} for reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                        copy_value = orig_value
                                except Exception as e:
                                    logger.warning(f'deepcopy(): {e}: Cannot find reference of type {orig_value.eClass.name} of reference {k.eClass.name}.{ref.name} in deepcopy memo, using original')
                                    copy_value = orig_value
                            v.eSet(ref.name, copy_value)
    copy._isset = InternalSet(self._isset)  # copy over the eIsSet configuration, otherwise all attributes are set due to this deepcopy
    return copy


setattr(EObject, '__copy__', clone)
setattr(EObject, 'clone', clone)
setattr(EObject, '__deepcopy__', deepcopy)
setattr(EObject, 'deepcopy', deepcopy)
